#include <stdio.h>
#include "Graph.h"
#include "GraphAdjList.h"
#include "Log.h"
#include "Mem.h"
#include "Common.h"
#include "Queue.h"

struct _Graph {
        GraphType options;
        GraphImpl *implementation;
};

typedef enum { WHITE, GRAY, BLACK } Color;

#define graph_is_weighted(g)      ((g)->options & GRAPH_TYPE_WEIGHTED)

Graph *graph_new (GraphType options, size_t number_of_vertices)
{
        Graph *graph = mem_new(Graph, 1);
        if (options & GRAPH_TYPE_DIRECTED) {
                printf("directed graph\n");
        } else {
                printf("undirected graph\n");
        }
        if (options & GRAPH_TYPE_DENSE) {
                printf("dense graph. Unsupported, aborting...\n");
                /* TODO use an implementation with adjacency matrices */
                exit(1);
        } else {
                printf("sparse graph\n");
                graph->implementation = graph_adj_list_new(options,number_of_vertices);
        }
        graph->options = options;
        return graph;
}

void graph_destroy (Graph *graph)
{
        log_return_if_fail (graph != NULL);
        graph->implementation->destroy(graph->implementation->private_data);
}

int graph_add_edge (Graph *graph, vertex from_a, vertex to_b, weight weight)
{
        log_return_val_if_fail (graph != NULL, FALSE);
        
        if (!graph_is_weighted(graph)) {
                log_return_val_if_fail (weight == 0, FALSE);
        }
        return graph->implementation->add_edge(graph->implementation->private_data, from_a, to_b, weight);
}

size_t graph_number_of_vertices (Graph *graph)
{
        log_return_val_if_fail (graph != NULL, 0);
        return graph->implementation->number_of_vertices(graph->implementation->private_data);
}

size_t graph_number_of_edges (Graph *graph)
{
        log_return_val_if_fail (graph != NULL, 0);
        return graph->implementation->number_of_edges(graph->implementation->private_data);
}

int graph_exists_edge (Graph *graph, vertex a, vertex b)
{
        log_return_val_if_fail (graph != NULL, FALSE);
        return graph->implementation->exists_edge(graph->implementation->private_data, a, b);
}

int graph_is_connected (Graph *graph)
{
        log_return_val_if_fail (graph != NULL, FALSE);
        return graph->implementation->is_connected(graph->implementation->private_data);
}

int graph_vertex_is_on_path (Graph *graph, vertex v,
                               vertex from_vertex, vertex to_vertex)
{
        log_return_val_if_fail (graph != NULL, FALSE);
        return graph->implementation->vertex_is_on_path(graph->implementation->private_data, v, from_vertex, to_vertex);
}

Graph** graph_get_components   (Graph *graph)
{
        /* TODO */
        return NULL;
}

SList *graph_get_neighbours     (Graph *graph, vertex v)
{
        /* TODO */
        return NULL;
}

void graph_foreach_neighbour (Graph *graph, vertex v, VisitorFunc visit, void *extra_data)
{
        /* TODO */
}

void graph_foreach_vertex (Graph *graph, VisitorFunc visit, void *extra_data)
{
        /* TODO */
}

void graph_foreach_edge (Graph *graph, TripletVisitor visit, void *extra_data)
{
        /* TODO */
}

void graph_dfs (Graph *graph, vertex initial, CancellableVisitor visit, void *extra_data)
{
        /* TODO */
}

void graph_bfs (Graph *graph, vertex initial, CancellableVisitor visit, void *extra_data)
{
        Color *color;
        Queue *queue;
        int number_of_vertices;
        vertex v;

        log_return_if_fail (graph != NULL);
        log_return_if_fail (visit != NULL);

        number_of_vertices = graph_number_of_vertices(graph);
        color = mem_new(Color, number_of_vertices);
        for (v = 0; v < number_of_vertices; v++) {
                if (v != initial) {
                        color[v] = WHITE;
                }
        }
        color[initial] = GRAY;
        queue = queue_new();

        /* TODO */
        
        queue_destroy(queue);
        mem_free(color);
}

SList *graph_topological_sort (Graph *graph)
{
       /* TODO */
       return NULL;
}
